from typing import List
from pytest import approx

from dinjo.model import ModelIVP, StateVariable, Parameter
from dinjo.predefined.epidemiology import (
    ModelSEIR, ModelSimpleSEIRV, ModelSIR, ModelSEIRV
)


class TestEpidemiologyModels:
    def setup(self):
        S = StateVariable('Succeptible', 'S', 0.5)
        E = StateVariable('Exposed', 'E', 0.5)
        Inf = StateVariable('Infected', 'I', 0.5)
        R = StateVariable('recovered', 'R', 0.5)
        V = StateVariable('Virus', 'V', 0.5)

        Lmbd = Parameter('Lambda', r'\Lambda', 1 / 2)
        mu = Parameter('mu', r'\mu', 1 / 2)
        inv_alpha = Parameter('inverse_alpha', r'\frac{1}{\alpha}', 1 / 2)
        omega = Parameter('omega', r'\omega', 1 / 2)
        gamma = Parameter('gamma', r'\gamma', 1 / 2)
        xi_E = Parameter('xi_E', r'\xi_E', 1 / 2)
        xi_I = Parameter('xi_I', r'\xi_I', 1 / 2)
        sigma = Parameter('sigma', r'\sigma', 1 / 2)
        beta_E = Parameter('beta_E', r'\beta_E', 1 / 2)
        beta_I = Parameter('beta_I', r'\beta_I', 1 / 2)
        beta_V = Parameter('beta_V', r'\beta_V', 1 / 2)
        c_E = Parameter('c_E', r'\c_E', 1 / 2)
        c_I = Parameter('c_I', r'\c_I', 1 / 2)
        c_V = Parameter('c_V', r'\c_V', 1 / 2)

        chi = Parameter('chi', r'\chi', 1 / 2)
        eta = Parameter('eta', r'\eta', 1 / 2)
        Pi = Parameter('Pi', r'\Pi', 1 / 2)
        tau = Parameter('tau', r'\tau', 1 / 2)

        self.sir_state_vars = [S, Inf, R]
        self.sir_params = [Lmbd, mu, omega, gamma, chi, eta, Pi, tau]

        self.seir_state_vars = [S, E, Inf, R]
        self.seir_params = [
            Lmbd, mu, omega, gamma, inv_alpha, chi, beta_E, beta_I
        ]

        self.seirv_state_vars = [S, E, Inf, R, V]
        self.seirv_params = [
            Lmbd, mu, inv_alpha, omega, gamma, xi_E, xi_I, sigma, beta_E,
            beta_I, beta_V, c_E, c_I, c_V
        ]

        self.seirv_simple_params = [
            Lmbd, mu, omega, gamma, inv_alpha, xi_E, xi_I, sigma,
            beta_E, beta_I, beta_V
        ]

        self.t_span = [0, 1]
        self.t_steps = 2

    def assert_approx_dydt(
        self,
        model: ModelIVP,
        approx_dydt: List[float],
        rel_tol: float = 0.99
    ):
        dydt = model.build_model(
            self.t_span[0],
            model.state_variables_init_vals,
            *model.parameters_init_vals
        )

        print(dydt)

        assert dydt == approx(approx_dydt, rel_tol)

    def test_seir(self):
        model_seir = ModelSEIR(
            self.seir_state_vars, self.seir_params, self.t_span, self.t_steps
        )
        self.assert_approx_dydt(model_seir, [-0.25, -0.25, -0.5, 0.25])

    def test_seirv(self):
        model_seirv = ModelSEIRV(
            self.seirv_state_vars, self.seirv_params, self.t_span, self.t_steps
        )
        self.assert_approx_dydt(model_seirv, [-0.05, -0.2, -0.5, 0.0, 0.25])

    def test_seirv_fixed(self):
        model_seirv_fixed = ModelSimpleSEIRV(
            self.seirv_state_vars, self.seirv_simple_params, self.t_span,
            self.t_steps
        )
        self.assert_approx_dydt(
            model_seirv_fixed,
            [-0.125, -0.125, -0.5, 0.0, 0.25]
        )

    def test_sir(self):
        model_sir = ModelSIR(
            self.sir_state_vars, self.sir_params, self.t_span, self.t_steps
        )
        self.assert_approx_dydt(model_sir, [0.125, -0.875, -0.25])



from typing import Any, List
import pytest

from numpy import pi
from numpy.random import random

from dinjo.model import StateVariable, Parameter
from dinjo.optimizer import Optimizer
from dinjo.predefined.physics import ModelOscillator


@pytest.fixture(scope='session')
def ho_state_vars():
    # Harmonic Oscillator Initial Value Problem
    q = StateVariable(
        name='position', representation='q', initial_value=1.0
    )
    p = StateVariable(
        name='momentum', representation='p', initial_value=0.0
    )
    return [q, p]


@pytest.fixture(scope='session')
def ho_params():
    # Define Paramters
    omega = Parameter(
        name='frequency', representation='w', initial_value=2 * pi, bounds=[4, 8]
    )
    return [omega]


@pytest.fixture(scope='session')
def t_span():
    return [0, 1]


@pytest.fixture(scope='session')
def t_steps():
    return 50


@pytest.fixture(scope='session')
def model_oscillator(ho_state_vars, ho_params, t_span, t_steps):
    # Instantiate the IVP class with appropiate State Variables and Parameters
    return ModelOscillator(
        state_variables=ho_state_vars,
        parameters=ho_params,
        t_span=t_span,
        t_steps=t_steps
    )


@pytest.fixture(scope='session')
def oscillator_solution(model_oscillator: ModelOscillator):
    return model_oscillator.run_model()


@pytest.fixture(scope='session')
def ho_mock_values(oscillator_solution: Any, t_steps: List[float]):
    noise_factor = 0.3
    return (
        oscillator_solution.y[0]
        + (2 * random(t_steps) - 1) * noise_factor
    )


@pytest.fixture(scope='session')
def ho_optimizer(
    model_oscillator: ModelOscillator,
    oscillator_solution: Any,
    ho_mock_values: List[float],
    ho_state_vars: List[StateVariable]
):
    return Optimizer(
        model_oscillator,
        ho_state_vars[0],
        ho_mock_values,
        oscillator_solution.t
    )



from typing import List

from ...model import ModelIVP


class ModelSEIRV(ModelIVP):

    def build_model(
        self, t, y,
        Lmbd, mu, inv_alpha, omega, gamma, xi_E, xi_I, sigma, beta_E, beta_I, beta_V, c_E, c_I, c_V
    ) -> List[float]:
        """
        Returns the vector field dy/dt evaluated at a given point in phase space
        """
        S, E, I, R, V = y

        def beta(x, b, c):
            return b / (1. + c * x)

        principal_flux = S * (
            beta(E, beta_E, c_E) * E + beta(I, beta_I, c_I) * I + beta(V, beta_V, c_V) * V
        )

        dydt = [
            Lmbd - principal_flux - S * mu,
            principal_flux - (inv_alpha + mu) * E,
            inv_alpha * E - (omega + gamma + mu) * I,
            gamma * I - mu * R,
            xi_E * E + xi_I * I - sigma * V
        ]

        return dydt



from ._seir_model import ModelSEIR
from ._seirv_model import ModelSEIRV
from ._seirv_fixed import ModelSimpleSEIRV
from ._sir_model import ModelSIR

__all__ = [
    'ModelSEIR',
    'ModelSEIRV',
    'ModelSimpleSEIRV',
    'ModelSIR',
]



from copy import deepcopy
from typing import Callable, List, Dict, Any

import math
import numpy as np
import scipy.optimize as opt

from . import model


class Optimizer:
    """Optimizes the initial value problem's parameters, as defined in
    the class :class:`ModelIVP`.

    Attributes
    ----------
    model : :class:`ModelIVP`
        the initial value problem to be optimized.
    reference_state_variable :class:`StateVariable`
        the state variable to be fitted to the solution of the IVP.
    reference_values : list[float]
        the 'experimental data' of the reference state variable to be
        used as the fitting variable.
    reference_t_values : list[float]
        the corresponding times at which the reference_values are given.
    integration_method : str
        must be one of the methods accepted by
        ``scipy.integrate.solve_ivp``.

    """
    def __init__(
        self,
        model: model.ModelIVP,
        reference_state_variable: model.StateVariable,
        reference_values: List[float],
        reference_t_values: List[float],
        integration_method: str = 'RK45',
    ) -> None:
        self.model = deepcopy(model)

        self.reference_state_variable = reference_state_variable
        self.reference_values = reference_values

        self.reference_t_values = reference_t_values

        self.integration_method = integration_method

        # Index of reference state variable in state variable list
        try:
            self._reference_state_variable_index: int = \
                model.state_variables.index(
                    reference_state_variable
                )
        except ValueError:
            raise ValueError(
                "self.reference_state_variable must be in model.state_variables"
            )

    @property
    def reference_t_values(self):
        return self._reference_t_values

    @reference_t_values.setter
    def reference_t_values(self, reference_t_values_input: List[float]):
        if len(self.reference_values) != len(reference_t_values_input):
            raise ValueError(
                "self.reference_values and self.reference_t_values must have the same length"
            )

        for i, t in enumerate(reference_t_values_input[:-1]):
            if not t < reference_t_values_input[i + 1]:
                raise ValueError(
                    "self.reference_t_values must be a list of floats in increasing order"
                )

        rel_tol = 0.999

        t_span_condition = \
            math.isclose(
                self.model.t_span[0], reference_t_values_input[0], rel_tol=rel_tol
            ) and math.isclose(
                self.model.t_span[1], reference_t_values_input[-1], rel_tol=rel_tol
            )

        if not t_span_condition:
            raise ValueError(
                "self.model.t_span and self.reference_t_values initial and "
                "final entries must coincide."
            )

        self._reference_t_values = reference_t_values_input

    def cost_function(
        self,
        parameters: List[float],
        cost_method: str = 'root_mean_square'
    ):
        """Function to be minimized by the optimizer.
        Initially this will be the root mean square of the difference between
        the observations and the numerical solution.

        Parameters
        ----------
        parameters : list[float]
            parameters of the model to be minimized. The order of the
            parameters must be the same as they appear in
            ``self.model.parameters``.
        cost_method : str
            Must be one of ``['root_mean_square',]``.
        """
        self.model.t_eval = self.reference_t_values
        solution = self.model.run_model(
            parameters=parameters, method=self.integration_method
        )

        # Get appropiate state variable solution in order to compare with
        # reference values
        solution_reference = solution.y[self._reference_state_variable_index]

        def rms(solution_reference) -> float:
            """Root mean square of the difference between ``reference_values``
            and ``solution_reference``.

            """
            diff_squared: np.ndarray = (
                (np.array(solution_reference) - np.array(self.reference_values)) ** 2
            )
            root_mean_square: float = np.mean(diff_squared) ** 0.5
            return root_mean_square

        cost_method_dict: Dict[str, Callable] = {
            'root_mean_square': rms,
        }

        try:
            cost_func = cost_method_dict[cost_method]
        except KeyError:
            raise ValueError(
                f"cost_method value '{cost_method}' not permitted. "
                f"Valid options options are {cost_method_dict.keys()}."
            )

        return cost_func(solution_reference)

    def optimize(
        self,
        cost_method: str = 'root_mean_square',
        algorithm: str = 'differential_evolution',
        algorithm_kwargs: Dict[str, Any] = {},
    ) -> opt.OptimizeResult:
        """Global minimization of cost function.

        Parameters
        ----------
        cost_function_method : str
            Must be one of the permitted values for cost_method parameter in
            :meth:`Optimize.cost_function`.
        algorithm : str
            scipy.optimize algorithm used for the optimization. Must be
            one of ``'differential_evolution'``, ``'shgo'``,
            ``'dual_annealing'``.
        algorithm_kwargs : dict[str, any]
            parameters passed to the optimization algorithm. They are
            different depending on the algorithm.

        Returns
        -------
        minimization
            object containing the minimization result as returned by
            the appropiate scipy algorithm (e.g. if the chosen al).
        """
        minimize_algorithms = {
            'differential_evolution': opt.differential_evolution,
            'shgo': opt.shgo,
            'dual_annealing': opt.dual_annealing,
        }

        try:
            minimize_algorithm = minimize_algorithms[algorithm]
        except KeyError:
            raise ValueError(
                "value of algorithm not permitted. Accepted values are "
                f"{minimize_algorithms.keys()}"
            )

        bounds = [
            param.bounds for param in self.model.parameters
        ]

        try:
            minimization = minimize_algorithm(
                func=self.cost_function,
                bounds=bounds,
                args=(cost_method,),
                **algorithm_kwargs
            )
        except TypeError:
            algorithm = [
                alg[0] for alg in minimize_algorithms.items()
                if alg[1] is minimize_algorithm
            ]
            algorithm = 'scipy.optimize.' + algorithm.pop(0)
            raise ValueError(
                'algorithm_kwargs value passed to minimize_global() got '
                f'unexpected keyword arguments for {algorithm}. Check '
                f'{algorithm} documentation for appropiate items in '
                'algorithm_kwargs.'
            )
        except RuntimeError:
            raise RuntimeError(
                'The minimization algorithm in minimize_global() got an '
                'internal error. Please check if the value of the kwarg '
                'cost_method passed to minimize_global() is a valid value for '
                'cost_method kwarg in Optimize.cost_function(). Please also '
                'check that your model was appropiately constructed.'
            )

        return minimization



"""Some helper functions and variables used to execute the examples of the
cmodels package.
"""
from datetime import datetime, timedelta
import os
import csv
from pathlib import Path

from typing import Any, List, Union
from numbers import Number


def param_sample_range(
    param: Union[int, float],
    param_range_width: Union[int, float]
):
    """Define the bounds of some parameters."""
    error_message = (
        "param_sample_range(): param and param_range width must be"
        "numbers greater than zero"
    )

    if not (
        isinstance(param, Number) and isinstance(param_range_width, Number)
    ):
        raise TypeError(error_message)

    if param < 0 or param_range_width < 0:
        raise ValueError(error_message)

    return [param / param_range_width, param * param_range_width]


def setup_csv(
    file_complete_path: str,
    first_row: List[Any] = None
):
    """If ``file_complete_path`` does not exist, creates a [CSV] file and adds
    a first row to it."""

    file_complete_path = Path(file_complete_path)

    file_parent_directory_path = file_complete_path.parent
    if not os.path.isdir(file_parent_directory_path):
        raise ValueError(
            'setup_csv(): file_complete_path parameter.'
            f'{file_parent_directory_path.parent} must be a directory.'
        )

    if os.path.isfile(file_complete_path):
        return None

    if first_row and (not isinstance(first_row, list) or isinstance(first_row, tuple)):
        raise TypeError(
            'setup_csv(): first_row kwarg must be a list or a tuple.'
        )

    with open(file_complete_path, 'w', newline='') as file:
        if first_row:
            csv_writer = csv.writer(file)
            csv_writer.writerow(first_row)

    return True


def int_to_str_date(i: int, initial_date: datetime) -> str:
    return (datetime.utcnow() + timedelta(days=i)).strftime('%Y-%m-%d')



from pytest import approx
from numpy import pi

from dinjo.model import StateVariable, Parameter
from dinjo.predefined.physics import ModelOscillator


def test_model_oscillator():
    q = StateVariable('position', 'q', 1.0)
    p = StateVariable('momentum', 'p', 0.0)
    omega = Parameter('frequency', 'w', 2 * pi, [4, 8])
    oscillator_model = ModelOscillator([q, p], [omega], [0, 1], 50)
    dydt = oscillator_model.build_model(0, [1/2, 1/2], 1/2)            # noqa
    assert dydt == approx([1/2, - 1/8], 0.99)                           # noqa



from ...model import ModelIVP


class ModelOscillator(ModelIVP):
    def build_model(self, t, y, w):
        """Harmonic Oscillator differential equations
        """
        q, p = y

        # Hamilton's equations
        dydt = [
            p,
            - (w ** 2) * q
        ]

        return dydt



import numbers
from typing import List, Optional

import numpy as np
from scipy.integrate import solve_ivp


class Variable:
    """Represents a variable

    Attributes
    ----------
    name : str
        name of the variable.
    representation : str
        string representing the state variable (could be the same as
        name.)
    initial_value : float
        reference value of the variable being represented.
    """
    def __init__(
        self, name: str, representation: str, initial_value: float = 0,
        *args, **kwargs
    ) -> None:
        self.name = name
        self.representation = representation
        self.initial_value = initial_value


class StateVariable(Variable):
    """Represents a State Variable of an initial value problem."""
    pass


class Parameter(Variable):
    """Represents a parameter of the differential equations definining
    an initial value problem.

    In addition to the attributes defined in
    :py:class:`dinjo.model.Variable`

    Attributes
    ----------
    bounds : 2-tuple of floats.
        list containing the minimum and maximum values that the
        parameter can take ``(min, max)``.
    """
    def __init__(
        self, name: str, representation: str, initial_value: float = 0,
        bounds: Optional[List[float]] = None, *args, **kwargs
    ) -> None:
        super().__init__(name, representation, initial_value)
        self.bounds = bounds if bounds else [initial_value, initial_value]

    @property
    def bounds(self):
        return self._bounds

    @bounds.setter
    def bounds(self, bounds_input):
        attr_err_message = "bounds must be a list of non decreasing numbers."
        init_val_not_in_bounds_range = "initial_value must be in the range defined by bounds."

        type_check: bool = (
            isinstance(bounds_input, list)
            and len(bounds_input) == 2
            and isinstance(bounds_input[0], numbers.Number)
            and isinstance(bounds_input[1], numbers.Number)
        )

        if not type_check:
            raise ValueError(attr_err_message)

        order_check: bool = bounds_input[0] <= bounds_input[1]

        if not order_check:
            raise ValueError(attr_err_message)

        if not (
            bounds_input[0] <= self.initial_value
            and bounds_input[1] >= self.initial_value
        ):
            raise ValueError(init_val_not_in_bounds_range)

        self._bounds = bounds_input


class ModelIVP:
    """Defines and integrates an initial value problem.

    Attributes
    ----------
    state_variables : list[:class:`StateVariable`]
    parameters : list[:class:`Parameter`]
    t_span : 2-tuple of floats
        Interval of integration (t0, tf). The solver starts with t=t0
        and integrates until it reaches t=tf.
    t_steps : int
        The solver will get the solution for ``t_steps`` equally
        separated times from ``t0`` to ``tf``.
    t_span : list[float]
        List containing the time values in which the user wants to
        evaluate the solution. All the values must be within the
        interval defined by ``t_span``.
    """
    def __init__(
        self,
        state_variables: List[StateVariable],
        parameters: List[Parameter],
        t_span: Optional[List[float]] = None,
        t_steps: int = 50,
        t_eval: Optional[List[float]] = None
    ) -> None:
        self.state_variables = state_variables
        self.parameters = parameters
        self.t_span = t_span if t_span else [0, 10]
        self.t_steps = t_steps
        self.t_eval = t_eval if t_eval else list(np.linspace(*t_span, t_steps))

    def _get_variable_init_vals(
        self, variables: List[Variable]
    ) -> List[float]:
        return [var.initial_value for var in variables]

    @property
    def state_variables_init_vals(self) -> List[float]:
        """Get the values of the model's state variables initial values
        in the order they are currently stored in ``self.state_variables``.
        """
        return self._get_variable_init_vals(self.state_variables)

    @property
    def parameters_init_vals(self):
        """Get the values of the model's parameters initial values in
        theorder they are currently stored in ``self.parameters``.
        """
        return self._get_variable_init_vals(self.parameters)

    def build_model(self, t, y, *args):
        """Defines the differntial equations of the model.

        Override this method so that it contains the differential
        equations of your IVP. The signature of the method must be
        ``build_model(self, t, y, *args)`` where ``t`` is the time,
        ``y`` is the state vector, and ``args`` are other parameters of
        the system.

        Parameters
        ----------
        t : float
            time at which the differential equation must be evaluated.
        y : list[float]
            state vector at which the differential must be evaluated.
        *args : any
            other parameters of the differential equation

        Returns
        -------
        The method must return the time derivative of the
        state vector evaluated at a given time.

        Note
        ----
        The parameters must be defined in the same order in which the
        parameters are stored in :attr:`ModelIVP.parameters`.

        Note
        ----
        The state variable vector must be defined in the same order
        in which the state variables are stored in
        :attr:`ModelIVP.state_variables`.

        Example
        -------
        For example if you want to simulate a harmonic
        oscillator of frequency :math:`\omega` and mass :math:`m`
        this method must be implemented as follows:

        .. code-block:: python

            def build_model(self, t, y, w, m):
                q, p = y

                # Hamilton's equations
                dydt = [
                    p,                      # dq/dt
                    - (w ** 2) * q          # dp/dt
                ]

                return dydt
        """
        raise NotImplementedError

    def run_model(
        self,
        parameters: List[float] = None,
        method: str = 'RK45'
    ):
        """Integrate model using ``scipy.integrate.solve_ivp``

        Parameters
        ----------
        parameters : list[float]
            List of the values of the paramters of the initial value
            problem.
        method : srt
            Integration method. Must be one of the methods accepted
            by ``scipy.integrate.solve_ivp``

        Returns
        -------
        Bunch object with the following fields defined (same return type
        as scipy.integrate.solve_ivp):
        t : ndarray, shape (n_points,)
            Time points.
        y : ndarray, shape (n, n_points)
            Values of the solution at t.
        sol : OdeSolution or None
            Found solution as OdeSolution instance; None if dense_output
            was set to False.
        t_events : list of ndarray or None
            Contains for each event type a list of arrays at which an
            event of that type event was detected. None if events was
            None.
        y_events : list of ndarray or None
            For each value of t_events, the corresponding value of the
            solution. None if events was None.
        nfev : int
            Number of evaluations of the right-hand side.
        njev : int
            Number of evaluations of the Jacobian.
        nlu : int
            Number of LU decompositions.
        status : int
            Reason for algorithm termination:
        """

        parameters = (
            parameters if parameters is not None else self.parameters_init_vals
        )

        parameters_permitted_types = (list, tuple, np.ndarray)
        parameters_type_is_permitted = False

        for permitted_type in parameters_permitted_types:
            parameters_type_is_permitted += isinstance(parameters, permitted_type)

        if not parameters_type_is_permitted:
            raise TypeError(
                "parameters must be a list, tuple or numpy.ndarray"
            )

        solution = solve_ivp(
            fun=self.build_model,
            t_span=self.t_span,
            y0=self.state_variables_init_vals,
            method=method,
            t_eval=self.t_eval,
            args=parameters
        )

        return solution



from dinjo.model import Parameter, StateVariable
from cmodel_examples_utilities import param_sample_range

# Instantiation of state variables and parameters

# In the next implementation (in which the model is built from scratch)
# StateVariable and Parameter instances initialization ought be a bit
# different, but the idea is the same

# The values of the parameters were obtained from Boris' Mathematica script
# (name, representation, initial_value)
seirv_state_variables_colombia_src = [
    ("suceptible", "S", 50372424),
    ("exposed", "Ex", 0),
    ("infected", "If", 1),
    ("recovered", "R", 0),
    ("vaccinated", "V", 0),
]

# List of instances of state variables
seirv_state_variables_colombia = [
    StateVariable(*sv) for sv in seirv_state_variables_colombia_src
]

# Parameters not variated by Boris' Mathematica script
Lmbd = 2083.62
mu = 0.0000150767
inv_alpha = 0.172414
omega = 0.0114
gamma = 0.0666667
xi_I = 6.32609
sigma = 1.0

# Optimal parameters found by Boris' Mathematica script
xi_E_opt = 1.09736
b_E_opt = 2.91823E-9
b_I_opt = 2.67472E-9
b_V_opt = 4.635E-9
c_E_opt = 0.0000855641
c_I_opt = 2.94906E-7
c_V_opt = 0.159107

# Parameters used to define their own bounds via the function
# param_sample_range
xi_E = 2.3
b_E, b_I, b_V = 1.5 * 3.11e-9, 1.5 * 0.62e-9, 1.5 * 1.03e-9

# Bounds of c and xi_I parameters used in Boris' Mathematica script
c_bounds = [0, 1]
xi_I_bounds = [0, 10]

# param_range_width used in Boris' Mathematica script
param_range_width = 3

# Each element of the list refers to a parameter
# (name, representation, initial_value, bounds)
seirv_parameters_colombia_src_ini = [
    ("Lambda", "Lmbd", Lmbd, [Lmbd, Lmbd]),
    ("mu", "mu", mu, [mu, mu]),
    ("inv_alpha", "inv_alpha", inv_alpha, [inv_alpha, inv_alpha]),
    ("omega", "omega", omega, [omega, omega]),
    ("gamma", "gamma", gamma, [gamma, gamma]),
    ("xi_E", "xi_E", xi_E_opt, param_sample_range(xi_E, param_range_width)),
    ("xi_I", "xi_I", xi_I, xi_I_bounds),
    ("sigma", "sigma", sigma, [sigma, sigma]),
    ("b_E", "b_E", b_E_opt, param_sample_range(b_E, param_range_width)),
    ("b_I", "b_I", b_I_opt, param_sample_range(b_I, param_range_width)),
    ("b_V", "b_V", b_V_opt, param_sample_range(b_V, param_range_width)),
    ("c_E", "c_E", c_E_opt, c_bounds),
    ("c_I", "c_I", c_I_opt, c_bounds),
    ("c_V", "c_V", c_V_opt, c_bounds),
]
seirv_parameters_colombia_src = [
    {
        'name': param_info[0],
        'representation': param_info[1],
        'initial_value': param_info[2],
        'bounds': param_info[3]
    }
    for param_info in seirv_parameters_colombia_src_ini
]

# List of instances of parameters
seirv_parameters_colombia = []
for param_info in seirv_parameters_colombia_src:
    seirv_parameters_colombia.append(
        Parameter(**param_info)
    )



import pytest

from dinjo.model import Variable, StateVariable, Parameter, ModelIVP
from dinjo.predefined.physics import ModelOscillator

VAR_PARAMS = 'foo', 'foo', 0.0


def test_variable():
    Variable(*VAR_PARAMS)
    assert True


def test_state_variable():
    StateVariable(*VAR_PARAMS)
    assert True


def test_parameter():
    Parameter(*VAR_PARAMS, bounds=[0.0, 1.0])


def test_parameter_bounds_not_explicit():
    Parameter(*VAR_PARAMS)


def test_get_parameter():
    p = Parameter(*VAR_PARAMS)
    p.bounds


def test_parameter_bounds_not_list():
    with pytest.raises(ValueError):
        Parameter(*VAR_PARAMS, bounds=(1.0, 0.0))


def test_parameter_bounds_not_number():
    with pytest.raises(ValueError):
        Parameter(*VAR_PARAMS, bounds=['a', 0.0])


def test_parameter_bounds_reversed():
    with pytest.raises(ValueError):
        Parameter(*VAR_PARAMS, bounds=[1.0, 0.0])


def test_parameter_init_val_not_in_bounds():
    with pytest.raises(ValueError):
        Parameter(*VAR_PARAMS, bounds=[0.1, 1.0])


class TestModelIVP:
    def setup(self):
        self.v = StateVariable(*VAR_PARAMS)
        self.p = Parameter(*VAR_PARAMS)
        self.model_ivp = ModelIVP(
            [self.v], [self.p], t_span=[0., 1.], t_steps=50
        )

    def test_state_vars_init_vals(self):
        sv_iv = self.model_ivp.state_variables_init_vals
        assert sv_iv == [self.v.initial_value]

    def test_parameters_init_vals(self):
        params_iv = self.model_ivp.state_variables_init_vals
        assert params_iv == [self.p.initial_value]

    def test_build_model_not_implemented(self):
        with pytest.raises(NotImplementedError):
            self.model_ivp.build_model(0, 0)

    def test_run_model(self, model_oscillator: ModelOscillator):
        model_oscillator.run_model()

    def test_run_model_params_not_list(
        self, model_oscillator: ModelOscillator
    ):
        with pytest.raises(TypeError):
            model_oscillator.run_model(parameters=set((0, 1)))



import os
from typing import Any, Dict, List, Union
from time import time

import matplotlib.pyplot as plt
import pandas as pd

from dinjo import model
from dinjo.predefined.epidemiology import ModelSEIRV
# State variables and parameters are setup in col_vars_params.py module
from col_vars_params import (
    seirv_state_variables_colombia,
    seirv_parameters_colombia
)


# Time span and time steps in days
t_span_col = [0, 171]
t_steps_col = 172

this_file_dir = os.path.dirname(__file__)
project_root_dir = os.path.join(this_file_dir, '..')

infected_reference_col_path = os.path.join(
    this_file_dir, '..', 'example_data', 'infected_reference_col.csv'
)

infected_reference_col = pd.read_csv(
    infected_reference_col_path
)['infected_reference_col'].to_numpy().T


def seirv_model_example(
    state_variables: List[model.StateVariable] = seirv_state_variables_colombia,
    parameters: List[model.Parameter] = seirv_parameters_colombia,
    t_span: List[Union[float, int]] = t_span_col,
    t_steps: int = t_steps_col,
    infected_reference: List[int] = infected_reference_col,
    monitor_computation_time: bool = False,
    plot_solution: bool = False,
) -> Dict[str, Any]:

    # Instantiate the model
    model_SEIRV = ModelSEIRV(
        state_variables=state_variables,
        parameters=parameters,
        t_span=t_span,
        t_steps=t_steps
    )

    # In the next iteration of cmodel's API the user will need to define the
    # model here manually by specifying the fluxes of each state variable using
    # some interface such as model_SEIRV.add_flux(...)

    # Run the model
    t0 = time()
    solution = model_SEIRV.run_model(method='RK45')
    if monitor_computation_time:
        print(f"Time calculating solution: {time() - t0}")

    if plot_solution:
        plt.figure()
        plt.plot(solution.t, solution.y[2], "ko-", label='SEIRV model')
        plt.plot(solution.t, infected_reference, "ro-", label='Data from INS')
        plt.xlabel('Days since case 1')
        plt.ylabel('Confirmed infected people')
        plt.legend()
        plt.grid()
        plt.tight_layout()
        plt.show()
        plt.close()

    seirv_model_example = {
        'model': model_SEIRV,
        'solution': solution
    }

    return seirv_model_example


if __name__ == "__main__":
    seirv_model_example(
        monitor_computation_time=True,
        plot_solution=True
    )



# Optimize the seirv model for colombia
# This script may take several minutes to execute
import os
import pickle
from time import time
from datetime import datetime
from typing import Any, Dict

import matplotlib.pyplot as plt
import pandas as pd
from pandas.errors import EmptyDataError

import dinjo.optimizer as optimizer
from seirv_model_colombia import (
    seirv_state_variables_colombia, seirv_model_example, infected_reference_col
)
from cmodel_examples_utilities import int_to_str_date, setup_csv


this_file_dir = os.path.dirname(__file__)
project_root_dir = os.path.join(this_file_dir, '..')


def optimizer_seirv_model_colombia_example(
    minimization_algorithm: str = 'differential_evolution',
    algorithm_kwargs={
        'popsize': 20,
        'disp': True,
        'tol': 0.0015,
        'maxiter': 100,
        'mutation': [0.3, 0.7],
        'atol': 200
    },
    print_optimization_log: bool = False,
    save_optimization_results: bool = False,
    generated_files_directory_name: str = 'generated_files',
    file_base_name: str = 'seirv_model_colombia_parameters_optimization',
    plot_results: bool = False,
    save_optimal_solution: bool = False,
    csv_optimal_solution_path: str = ''
) -> Dict[str, Any]:

    seirv_model_example_col = seirv_model_example()
    optimizer_seirv_model_colombia = optimizer.Optimizer(
        model=seirv_model_example_col["model"],
        reference_state_variable=seirv_state_variables_colombia[2],
        reference_values=infected_reference_col,
        reference_t_values=seirv_model_example_col['solution'].t
    )

    # Optimization using the desired algorith
    t0 = time()
    seirv_colombia_parameters_optimization = \
        optimizer_seirv_model_colombia.optimize(
            cost_method='root_mean_square',
            algorithm=minimization_algorithm,
            algorithm_kwargs=algorithm_kwargs
        )
    computation_time = time() - t0

    line_split = '=============================================\n'

    # Construct log message
    optimization_log = (
        line_split
        + 'SEIRV Model Colombia Parematers Optimization:\n'
        + line_split
        + f'Computation time: {computation_time}\n'
        + line_split
        + str(seirv_colombia_parameters_optimization)
        + '\n\nParam \tOptimized \tExpected\n'
        + line_split
    )

    optimal_solution_mathematica = \
        optimizer_seirv_model_colombia.model.parameters_init_vals
    parameters = optimizer_seirv_model_colombia.model.parameters

    for i, param in enumerate(seirv_colombia_parameters_optimization.x):
        optimization_log += (
            f"{parameters[i].representation}\t"
            f"{param:.4e}\t"
            f"{optimal_solution_mathematica[i]:.4e}\n"
        )

    if print_optimization_log:
        print(optimization_log)

    if save_optimization_results:
        # Directory name (full path). Created in the same folder as this file.
        generated_files_directory_path = os.path.join(
            this_file_dir,
            generated_files_directory_name
        )
        # File base name for pickle and txt (full path).
        file_base_path = os.path.join(
            generated_files_directory_path,
            file_base_name
        )

        # If directory does not exist, create it
        if not os.path.isdir(generated_files_directory_path):
            try:
                os.mkdir(generated_files_directory_path)
            except OSError:
                print(
                    "Creation of the directory "
                    f"{generated_files_directory_path} failed"
                )
            else:
                print(
                    "Successfully created the directory "
                    f"{generated_files_directory_path}"
                )

        csv_file_path = file_base_path + '.csv'

        # Create CSV file with proper column names if one does not exist
        param_names = [
            param.name for param in optimizer_seirv_model_colombia.model.parameters
        ]
        csv_column_names = [
            'index',
            'comp_time', 'algorithm',
            'success', 'optimizer_message', 'fun', 'nfev', 'nit',
            *param_names
        ]
        setup_csv(csv_file_path, first_row=csv_column_names)

        # Organize optimal parameters in dictionary
        optimal_parameters_dict = {}
        for i, param_opt_value in enumerate(seirv_colombia_parameters_optimization.x):
            optimal_parameters_dict[param_names[i]] = param_opt_value

        # Add optimal parameters to the CSV file using pandas
        try:
            df_optimization_log = pd.read_csv(
                csv_file_path, index_col=csv_column_names[0]
            )
        except (EmptyDataError, ValueError):
            df_optimization_log = pd.DataFrame({csv_column_names[1]: [0.]})

        optimization_csv_row = {
            csv_column_names[1]: computation_time,
            csv_column_names[2]: minimization_algorithm,
            csv_column_names[3]: seirv_colombia_parameters_optimization.success,
            csv_column_names[4]: seirv_colombia_parameters_optimization.message,
            csv_column_names[5]: seirv_colombia_parameters_optimization.fun,
            csv_column_names[6]: seirv_colombia_parameters_optimization.nfev,
            csv_column_names[7]: seirv_colombia_parameters_optimization.nit,
            **optimal_parameters_dict
        }
        df_optimization_log = df_optimization_log.append(
            optimization_csv_row, ignore_index=True
        )
        df_optimization_log.to_csv(csv_file_path, index_label=csv_column_names[0])

        # Save pickle
        pickle.dump(
            seirv_colombia_parameters_optimization,
            open(file_base_path + '.pickle', 'wb')
        )

    if plot_results or save_optimal_solution:
        optimal_solution = optimizer_seirv_model_colombia.model.run_model(
            parameters=seirv_colombia_parameters_optimization.x
        )

    if plot_results:

        plt.figure()
        plt.plot(
            optimal_solution.t, optimal_solution.y[2],
            "ko-", label='Optimization of SEIRV model'
        )
        plt.plot(
            optimal_solution.t, infected_reference_col,
            "ro-", label='Data from INS'
        )
        plt.xlabel('Days since case 1')
        plt.ylabel('Confirmed infected people')
        plt.legend()
        plt.grid()
        plt.tight_layout()
        plt.show()
        plt.close()

    optimizer_seirv_model_colombia_example = {
        'optimizer': optimizer_seirv_model_colombia,
        'optimization_results': seirv_colombia_parameters_optimization,
        'optimization_log': optimization_log,
        'computation_time': computation_time,
    }

    if save_optimal_solution:
        t_dates = []
        initial_date = datetime.utcnow()

        for i in range(len(optimal_solution.t)):
            t_dates.append(int_to_str_date(i, initial_date))

        df = pd.DataFrame(
            {
                'date': t_dates,
                'I_reference': infected_reference_col,
                'S_optimal': optimal_solution.y[0],
                'E_optimal': optimal_solution.y[1],
                'I_optimal': optimal_solution.y[2],
                'R_optimal': optimal_solution.y[3],
                'V_optimal': optimal_solution.y[4],
            }
        )

        df.set_index('date', inplace=True)

        df.to_csv(csv_optimal_solution_path)

    return optimizer_seirv_model_colombia_example


if __name__ == "__main__":
    csv_optimal_solution_path = os.path.join(
        this_file_dir,
        'generated_files',
        'seirv_model_colombia_optimal_solution.csv'
    )

    optimizer_seirv_model_colombia_example(
        minimization_algorithm='differential_evolution',
        print_optimization_log=True,
        save_optimization_results=True,
        plot_results=True,
        save_optimal_solution=False,
        csv_optimal_solution_path=csv_optimal_solution_path
    )



import os
from typing import Any, Dict, List, Union
import pickle
from datetime import datetime, timedelta


import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from dinjo import model, optimizer


this_file_dir = os.path.dirname(__file__)
project_root_dir = os.path.join(this_file_dir, '..')


class ModelOscillator(model.ModelIVP):
    def build_model(self, t, y, w):
        """Harmonic Oscillator differential equations
        """
        q, p = y

        # Hamilton's equations
        dydt = [
            p,
            - (w ** 2) * q
        ]

        return dydt


def oscillator_optimizer_example(
    q0: Union[float, int] = 1.0,
    p0: Union[float, int] = 0.0,
    w0: Union[float, int] = 2 * np.pi,
    w_bounds: List[Union[float, int]] = [4, 8],
    t_span: List[Union[float, int]] = [0, 1],
    t_steps: int = 50,
    fake_data_noise_factor: float = 0.30,
    minimization_algorithm: str = 'differential_evolution',
    save_optimization_results_pickle: bool = False,
    plot_solution: bool = False,
    save_csv_file: bool = False,
    date_fmt: bool = False
) -> Dict[str, Any]:

    # Define State Variables
    q = model.StateVariable(
        name='position', representation='q', initial_value=q0
    )
    p = model.StateVariable(
        name='momentum', representation='p', initial_value=p0
    )

    # Define Paramters
    omega = model.Parameter(
        name='frequency', representation='w', initial_value=w0, bounds=w_bounds
    )

    # Instantiate Model
    oscillator_model = ModelOscillator(
        state_variables=[q, p],
        parameters=[omega],
        t_span=t_span,
        t_steps=t_steps
    )

    # Run the model
    oscillator_solution = oscillator_model.run_model()

    # Build fake observation data from the solution (to test optimizer)
    oscillator_fake_position_data = (
        oscillator_solution.y[0]
        + (2 * np.random.random(t_steps) - 1) * fake_data_noise_factor
    )

    # Optimizer using fake noisy data
    oscillator_optimizer = optimizer.Optimizer(
        model=oscillator_model,
        reference_state_variable=q,
        reference_values=oscillator_fake_position_data,
        reference_t_values=oscillator_solution.t
    )

    # Optimize parameters
    oscillator_parameters_optimization = \
        oscillator_optimizer.optimize(algorithm=minimization_algorithm)

    date_equivalent = [
        datetime.now() + timedelta(days=i) for i in range(len(oscillator_solution.t))
    ]

    # Save pickle
    if save_optimization_results_pickle or save_csv_file:
        files_directory_path = os.path.join(
            this_file_dir,
            'generated_files'
        )

        if not os.path.isdir(files_directory_path):
            try:
                os.mkdir(files_directory_path)
            except OSError:
                print("Creation of the directory %s failed" % files_directory_path)
            else:
                print("Successfully created the directory %s " % files_directory_path)

                if save_optimization_results_pickle:
                    pickle_path = os.path.join(
                        files_directory_path,
                        'oscillator_parameters_optimization.pickle'
                    )
                    pickle.dump(
                        oscillator_parameters_optimization,
                        open(pickle_path, 'wb')
                    )

    # Calculate differential equation solution using optimized parameters
    if oscillator_parameters_optimization.success:
        oscillator_optimal_solution = oscillator_model.run_model(
            parameters=oscillator_parameters_optimization.x
        )
        if save_csv_file:
            csv_path = os.path.join(
                files_directory_path,
                'oscillator_csv_example.csv'
            )
            pd.DataFrame(
                {
                    'date': [
                        date.strftime('%Y-%m-%d') for date in date_equivalent
                    ],
                    'user_data_variable_X': oscillator_fake_position_data,
                    'optimal_solution_variable_X': oscillator_optimal_solution.y[0]
                }
            ).to_csv(csv_path, header=True, index=False)
    else:
        oscillator_optimal_solution = None
        print("Parameter optimization did not succed.")

    # Plot solution
    if plot_solution:
        oscillator_solution.t = date_equivalent if date_fmt else oscillator_solution.t
        oscillator_optimal_solution.t = date_equivalent if date_fmt else oscillator_solution.t
        xdate_val = True if date_fmt else False

        plt.figure()
        if date_fmt:
            plt.xticks(rotation=70)
        plt.plot_date(
            oscillator_solution.t, oscillator_solution.y[0],
            'k-',
            label='Exact Solution using '
                  f'$\omega={omega.initial_value:.3f}$',
            xdate=xdate_val
        )
        plt.plot_date(
            oscillator_solution.t, oscillator_fake_position_data,
            'ro', label='Noisy fake data', xdate=xdate_val
        )
        try:
            plt.plot_date(
                oscillator_optimal_solution.t, oscillator_optimal_solution.y[0],
                'k-*',
                label='Optimized solution from noisy data\n'
                      f'using {minimization_algorithm} algorithm\n'
                      f'$\omega={oscillator_parameters_optimization.x[0]:.3f}$',
                xdate=xdate_val
            )
        except Exception:
            pass
        plt.xlabel('t')
        plt.ylabel('q(t)')
        plt.legend()
        plt.grid()
        plt.tight_layout()
        plt.show()
        plt.close()

    oscillator_optimization_example = {
        'model': oscillator_model,
        'solution': oscillator_solution,
        'fake_position_data': oscillator_fake_position_data,
        'optimizer': oscillator_optimizer,
        'parameters_optimization': oscillator_parameters_optimization,
        'optimal_solution': oscillator_optimal_solution,
    }

    return oscillator_optimization_example


if __name__ == '__main__':
    oscillator_optimizer_example(
        plot_solution=True,
        save_optimization_results_pickle=True,
        save_csv_file=True,
        date_fmt=False
    )



from typing import Any, List

import pytest

from dinjo.model import ModelIVP, Parameter, StateVariable
from dinjo.optimizer import Optimizer
from dinjo.predefined.physics import ModelOscillator


def test_bad_reference_state_variable(
    model_oscillator: ModelOscillator,
    oscillator_solution: Any,
    ho_mock_values: List[float],
):
    with pytest.raises(ValueError):
        Optimizer(
            model_oscillator,
            StateVariable('foo', 'foo', 0.0),
            ho_mock_values,
            oscillator_solution.t
        )


def test_reference_values_bad_length(
    model_oscillator: ModelOscillator,
    ho_state_vars: List[StateVariable],
    ho_mock_values: List[float],
    oscillator_solution: Any
):
    longer_ref_values = list(ho_mock_values[:])
    longer_ref_values.append(0)
    with pytest.raises(ValueError):
        Optimizer(
            model_oscillator,
            ho_state_vars[0],
            longer_ref_values,
            oscillator_solution.t
        )


def test_reference_t_values_bad_length(
    model_oscillator: ModelOscillator,
    ho_state_vars: List[StateVariable],
    ho_mock_values: List[float],
    oscillator_solution: Any
):
    longer_t_vlaues = list(oscillator_solution.t[:])
    longer_t_vlaues.append(0)
    with pytest.raises(ValueError):
        Optimizer(
            model_oscillator,
            ho_state_vars[0],
            ho_mock_values,
            longer_t_vlaues
        )


def test_unordered_reference_t_values(
    model_oscillator: ModelOscillator,
    ho_state_vars: List[StateVariable],
    ho_mock_values: List[float],
    oscillator_solution: Any
):
    longer_t_vlaues = list(oscillator_solution.t[:])
    longer_t_vlaues[1] = 10
    with pytest.raises(ValueError):
        Optimizer(
            model_oscillator,
            ho_state_vars[0],
            ho_mock_values,
            longer_t_vlaues
        )


def test_reference_t_values_bad_limits(
    model_oscillator: ModelOscillator,
    ho_state_vars: List[StateVariable],
    ho_mock_values: List[float],
    oscillator_solution: Any
):
    longer_t_vlaues = list(oscillator_solution.t[:])
    longer_t_vlaues[0] = 0.01
    with pytest.raises(ValueError):
        Optimizer(
            model_oscillator,
            ho_state_vars[0],
            ho_mock_values,
            longer_t_vlaues
        )


def test_reference_t_values_property(ho_optimizer: Optimizer):
    assert ho_optimizer.reference_t_values.any()


def test_cost_function_root_mean_square(
    model_oscillator: ModelOscillator,
    ho_optimizer: Optimizer
):
    ho_optimizer.cost_function(
        model_oscillator.parameters_init_vals,
        cost_method='root_mean_square'
    )


def test_cost_function_bad_cost_method(
    model_oscillator: ModelOscillator,
    ho_optimizer: Optimizer
):
    with pytest.raises(ValueError):
        ho_optimizer.cost_function(
            model_oscillator.parameters_init_vals,
            cost_method='foo'
        )


def test_optimize_algorithm_diff_evol(
    ho_optimizer: Optimizer
):
    ho_optimization = ho_optimizer.optimize(
        algorithm='differential_evolution'
    )
    assert ho_optimization.success


def test_optimize_algorithm_shgo(
    ho_optimizer: Optimizer
):
    ho_optimization = ho_optimizer.optimize(
        algorithm='shgo'
    )
    assert ho_optimization.success


@pytest.mark.skip('Takes more time. Comment this line to run this test.')
def test_optimize_algorithm_dual_annealing(
    ho_optimizer: Optimizer
):
    ho_optimization = ho_optimizer.optimize(
        algorithm='dual_annealing'
    )
    assert ho_optimization.success


def test_optimize_bad_algorithm_name(
    ho_optimizer: Optimizer
):
    with pytest.raises(ValueError):
        ho_optimizer.optimize(
            algorithm='foo'
        )


def test_optimize_bad_algotrithm_keyword_arguments(
    ho_optimizer: Optimizer
):
    with pytest.raises(ValueError):
        ho_optimizer.optimize(
            algorithm_kwargs={'foo': 'foo'}
        )


def test_optimize_runtime_error():
    class Foo(ModelIVP):
        def build_model(self, t, y, foo):
            return None
    variable_params = ('foo', 'foo', 0.0)
    foo_var = StateVariable(*variable_params)
    foo_param = Parameter(*variable_params)
    foo_model = Foo([foo_var], [foo_param], [0, 1], 2)
    foo_optimizer = Optimizer(foo_model, foo_var, [0, 1], [0, 1])
    with pytest.raises(RuntimeError):
        foo_optimizer.optimize()



from typing import List

from ...model import ModelIVP


class ModelSEIR(ModelIVP):

    def build_model(
        self, t, y,
        Lmbd, mu, omega, gamma, inv_alpha, chi, beta_E, beta_I
    ) -> List[float]:
        """
        Returns the vector field dy/dt evaluated at a given point in phase space
        """
        S, E, I, R = y

        dydt = [
            Lmbd - beta_E * S * E - beta_I * S * I - (chi + mu) * S,
            beta_E * S * E + beta_I * S * I - inv_alpha * E - mu * E,
            inv_alpha * E - (gamma + omega + mu) * I,
            gamma * I + chi * S - mu * R
        ]

        return dydt



from typing import List

from ...model import ModelIVP


class ModelSIR(ModelIVP):

    def build_model(
        self, t, y,
        Lmbd, mu, omega, gamma, chi, eta, Pi, tau
    ) -> List[float]:
        """
        Returns the vector field dy/dt evaluated at a given point in phase space
        """
        S, I, R = y

        dydt = [
            Lmbd + Pi * R - tau * S * I - (chi + mu) * S,
            tau * S * I - eta * R - (gamma + omega + mu) * I,
            gamma * I + chi * S - (eta + Pi + mu) * R
        ]

        return dydt
